//=============================================================================
//MrFu_Core.js
//=============================================================================
/*:de
* @plugindesc v1.1 Dieses Plugin wird für andere Plugins von mir benötigt.
*
* @param 
* @desc 
* @default 
*
* @author Mr. Fu
*
* @help 
*
* Dieses Plugin wird für andere Plugins von mir benötigt. Es enthält
* allgemeine Funktionen, Parameter und Variablen. Wenn ein anderes Plugin
* dieses benötigt, steht das in der Beschreibung. Sicherheitshalber sollte
* dieses Plugin oberhalb anderer Plugins von mir im Plug-in-Manager aufgeführt
* sein.
*
* Vielen Dank an dieser Stelle an waynee95 vom deutschen RPG-Maker-MV-Forum für
* seine Hilfe.
* https://rpgmaker-mv.de/user/7-waynee95/
*
* Folgende Funktionen sind im Plugin enthalten und können vom Programmierer
* selbst aufgerufen werden. Diese eignen sich allerdings fast ausschließlich
* zur Nutzung in anderen Plugins.
*
* $MFUIncludes(array, begriff, bezeichner):
* Diese Funktion prüft ob eine Eigenschaft eines Objektes in einem Array gleich
* einem Begriff ist und gibt dann entweder true oder false zurück. Falls true
* zurückgegeben wird, wird der globalen Variable '$MFULastIndex' die
* entsprechende Posistion des Objekts in dem Array mitgegeben.
* Folgende Argumente werden erwartet:
* 'array': Das zu durchsuchende Array. Es sollte nur aus Objekten bestehen
* 'begriff': Der zu suchende Inhalt.
* 'bezeichner': Der Name der Eigenschaft der Objekte des Arrays. Falls dieses
*               Argument frei gelassen wird, wird nach Eigenschaften mit dem
*               Namen 'bezeichnung' gesucht.
* Wenn also beispielsweise für die vierte Position in dem Array true
* zurückgegeben wird, wird $MFULastIndex auf 3 gesetzt. Diese Funktion
* gibt nur den ersten Treffer zurück!
*
* $MFUIsNumber(n):
* Diese Funktion überprüft, ob das Argument eine richtige Zahl ist und gibt
* entweder true oder false zurück.
*
* $MFUConsoleError():
* Diese Funktion gibt eine Alarmmeldung aus in der darauf verwiesen wird die
* Konsole zu öffnen.
*
* $MFUParseParameters(params):
* Diese Funktion wandelt Pluginparameter in nutzbare Variablen um. Sie
* erwartet als Argument 'params' die Parameter des jeweiligen Plugins und
* speichert sie als Array.
* Bsp.:
* var params = $MFUParseParameters(PluginManager.parameters("MrFu_Core"));
* Jetzt kann man wie gewohnt auf die Parameter zugreifen.
* Bsp.: var test = params['Hier den Namen einfügen'];
*
* $MFUCheckParams(prüfwert, name, alternative, typ):
* Diese Funktion prüft ob der Prüfwert entweder eine Zahl zwischen 1 und 5000
* ist, um z.B. korrekte Variablen- bzw. Schalter-IDs zu erhalten, oder ob er
* true bzw. false ist. Gedacht ist diese Funktion um die Pluginparameter auf
* Richtigkeit zu prüfen. Es wird eine entsprechende Meldung in der Konsole
* ausgegeben, falls die Werte nicht in Ordnung sind.
* Folgende Argumente werden erwartet:
* 'prüfwert': Der Parameter, der geprüft werden soll.
* 'name': Der Name des entsprechenden Parameters, damit er in der möglichen
*         Meldung erscheint.
* 'alternative': Der Wert, der in der Variable gespeichert werden soll, falls
*                der Prüfwert nicht in Ordnung ist.
* 'tpy': Hier entweder "zahl" oder "bool" eintragen, damit die Funktion weiß,
*        auf welchen Wert geprüft werden soll.
* Bsp.:
* var id = $MFUCheckParams(params['ID'], 'ID', 1, "zahl");
* var switch = $MFUCheckParams(params['Switch'], 'Switch', true, "bool");
*
* $MFURomanNumerals(zahl):
* Diese Funktion wandelt eine Zahl in eine römische Zahl um und gibt sie
* zurück. Sie erwartet als Argument die umzuwandelnde Zahl.
* Bsp.: var römisch = $MFURomanNumerals(1234)
* Gibt "MCXXXIV" zurück.
*
* $MFUSort(array, bezeichner):
* Diese Funktion sortiert ein Array alphabetisch nach der Eigenschaft die im
* Argument 'bezeichner' steht. Folgende Argumente werden erwartet:
* 'array': Das zu sortierende Array. Es darf nur aus Objekten bestehen.
* 'bezeichner': Der Name der Eigenschaft der Objekte des Arrays. Falls dieses
*               Argument frei gelassen wird, wird nach Eigenschaften mit dem
*               Namen 'bezeichnung' sortiert.
*
* $MFUShowMessageWindow(text, hint, pos):
* Diese Funktion zeigt ein einfaches Textfenster an. Folgende Argumente werden
* erwartet:
* 'text': Der anzuzeigende Text. Hier kann man natürlich auch die Funktion zum
*         Zeilenumbruch eintragen.
* 'hint': Der Wert für den Hintergrund. Entweder 0, 1 oder 2. Falls der Wert
*         frei gelassen wird, wird 0 gewählt.
*         0: ist der normale Textboxhintergrund
*         1: ist ein gedimmter Hintergrund
*         2: ist kein Hintergrund
* 'pos': Die Position des Textfensters. Entweder 0, 1 oder 2. Falls der Wert
*        frei gelassen wird, wird 2 gewählt.
*        0: Fenster ist oben 
*        1: Mittiges Fenster
*        2: Fenster ist unten
* 'face': Der Name der PNG-Datei, die das Faceset etnhält. Bsp.: "People1" (lädt
*         die People1.png im Pfad: img/faces)
* 'faceIndex': Eine Zahl die angibt, welches Face aus dem Faceset verwendet
*              werden soll. 0 = links oben, 3 = rechts oben, 4 = links unten,
*              7 = rechts unten
* 'scroll': true eingeben, falls es ein Scrolltext werden soll.
* 'noFast': true eingeben, falls der Scrolltext nicht beschleunigt werden darf.
* 'interpreter': this eingeben, falls das Spiel auf eine Eingabe warten soll.
* Bsp.:
* $MFUShowMessageWindow($MFUWordwrapping("Hallo.", undefined, 5*36), 1, 0,
*                       "People1", 2, false, false, this)
* 
* $MFURollDice(würfel):
* Diese Funktion ermittelt eine Ganzzahl zwischen 1 und dem Argument 'würfel',
* gibt sie zurück und zeigt das Ergebnis in der Konsole an.
* Bsp.: var w20 = $MFURollDice(20);
*
* $MFUNumberIntoWord(zahl):
* Diese Funktion wandelt Zahlen zwischen 0 und 12 in Wörter um und gibt sie
* zurück. Andere Zahlen bleiben unverändert. Sie erwartet als Argument die
* umzuwandelnde Zahl.
* Bsp.: var wort = $MFUNumberIntoWord(10)
* Gibt "zehn" zurück.
*
* $MFUWordwrapping(text, breite, höhe, keineMeldung, keinePrüfung):
* Diese Funktion fügt einem beliebigen Text Zeilenumbrüche hinzu. Folgende
* Argumente werden erwartet:
* 'text': Der jeweilige Text.
* 'breite': Die Breite in Pixel des jeweiligen Fensters. Falls keine Breite
*           übergeben wird, nimmt die Funktion die Breite des Spielfensters
*           (SceneManager._screenWidth).
* 'höhe': Die Höhe in Pixel des jeweiligen Fensters. Falls kein Höhe übergeben
*         wird, nimmt die Funktion die Höhe des Spielfensters
*         (SceneManager._screenHeight).
* 'keineMeldung': Falls in der Konsole keine Meldung ausgeworfen werden soll,
*                 hier 'true' (ohne die ') eintragen.
* 'keinePrüfung': Falls die Prüfung der Texthöhe nicht erfolgen soll, hier
*                 'true' (ohne die ') eintragen.
* Zeilenumbrüche werden eingefügt wenn die Zeile länger ist als das Fenster
* breit oder wenn ein Zeilenumbruch im Text enthalten ist. Dabei wird das Wort
* am Ende der Zeile nicht abgeschnitten, außer wenn es die ganze Zeile
* ausfüllt. Bindestriche werden erkannt aber nicht automatisch eingefügt!
* Sollte der Text mehr Zeilen haben als Platz durch die Höhe verfügbar ist,
* wird die Schriftgröße so oft verkleinert bis der Platz ausreicht oder die 
* Schriftgröße eine Stufe kleiner als die Standardgröße ist. Sollten es dann 
* immer noch mehr Zeilen sein als Platz zur Verfügung ist, wird eine
* Fehlermeldung ausgeworfen. Durch das Argument 'keinePrüfung' kann diese
* Prüfung übersprungen werden.
* ACHTUNG: Sollte in einem Text mehrmals die Schriftgröße verändert werden,
* kann es sein, dass die Funktion eine Fehlermeldung auswirft, obwohl genügend
* Platz verfügbar ist! Für diesen Fall ist das Argument 'keineMeldung' gedacht.
* In jedem Fall wird in die globalen Variablen '$MFULastTextHeight' die Höhe
* des Textes in Pixel gespeichert und in '$MFULastTextLength' die Länge des
* Textes in Pixel gespeichert
* Folgende Steuerzeichen können im Text verwendet werden:
* - \\v[x] Zeigt die X-te Variable an
* - \\n[x] Zeigt den Namen des X-ten Akteurs an
* - \\p[x] Zeigt den Namen des X-ten Gruppenmitglieds an
* - \\g    Zeigt die Bezeichnung der Währung an
* - \\c[x] Ändert nachfolgenden Text in die X-te Farbe
* - \\i[x] Zeigt das X-te Symbol an
* - \\{    Vergößert den nachfolgenden Text um eine Stufe (maximal sechs Mal ab
*          Standardgröße möglich)
* - \\}    Verkleinert den nachfolgenden Text um eine Stufe (maximal ein Mal ab
*          Standardgröße möglich)
* - \n     Manueller Zeilenumbruch (in diesem Fall nur ein \ voranstellen!)
* Alle anderen Steuerzeichen sind nicht möglich!
* Bsp.: text = $MFUWordwrapping(text, undefined, 512), falls keine Breite
*       übergeben werden soll.
*       text = $MFUWordwrapping(text, 816), falls keine Höhe übergeben werden
*       soll.
*       text = $MFUWordwrapping(text), falls nur der Text übergeben werden
*       soll.
*
* $MFUGainPropertyArray(array, eigenschaft)
* Diese Funktion gibt ein Array zurück, das aus Eigenschaften mehrerer Objekte
* besteht. Folgende Argumente werden erwartet:
* 'array': Ein Array welches die Objekte. Jedes der Objekte muss die gesuchte
*          Eigenschaft enthalten.
* 'eigenschaft': Die Eigenschaft der Objekte, die in das neue Array gespeichert
*                werden soll.
*
* ============================================================================
* Terms of Use
* ============================================================================
* Dieses Plugin darf frei genutzt werden. Credits sind erforderlich. Das Plugin
* darf nicht verändert werden.
*
* Kontakt: Mr. Fu auf www.rpgmaker-mv.de
* https://rpgmaker-mv.de/user/319-mr-fu/
* ============================================================================
* Changelog
* ============================================================================
* Version 1.1: 16.10.2021
* - Funktion '$MFUWordwrapping' um Argument 'keinePrüfung' erweitert
* - Funktion '$MFUShowMessageWindow' um Faces, Scolltext und Waitmode erweitert
* - Funktion '$MFUGainPropertyArray' erstellt
*
* Version 1.0: 11.07.2020
* - Release!
*/

const $MFUCoreVersion = 1.1;

(function() {

$MFUParseParameters = function(params) {
	var obj;
	try {
		obj = JsonEx.parse(params && typeof params === 'object' ? JsonEx.stringify(params) : params);
	} catch (e) {
				return params;
	} // end of try
	if (obj && typeof obj === 'object') {
		Object.keys(obj).forEach(function (key) {
			obj[key] = $MFUParseParameters(obj[key]); 
			// If the parameter has no value, meaning it's an empty string, just set it to null
			if (obj[key] === '') {
				obj[key] = null;
			} // end of if
		});
	} // end of if
	return obj;
} // end of ParseParameters

$MFUAdvice = false;

$MFUCheckParams = function(prüfwert, name, alternative, typ){
	if (typ == "zahl"){
		prüfwert = Number(prüfwert);
		if (!$MFUIsNumber(prüfwert) || prüfwert < 1 || prüfwert > 5000) {
			console.error("Der Eintrag im Parameter '%s' ist keine Zahl zwischen 1 und 5000. Er wird vorläufig auf '%d' gesetzt.", name, alternative);
			prüfwert = alternative;
			$MFUAdvice = true;
		} // end of if
	} else if (typ == "bool"){
		if (prüfwert === true || prüfwert === false) {

		} else {
			prüfwert = alternative;
		} // end of if
	} // end of if
	return prüfwert;
} // end of CheckParams

// Gibt das Würfelergebnis aus
$MFURollDice = function(würfel) { // Gibt eine Ganzzahl zwischen 1 und "würfel" zurück
	var zufall = Math.round(Math.random()*würfel+0.5);
	console.log("W" + würfel + " = " + zufall);
	return zufall;
} // end of RollDice

$MFULastTextHeight = 0;
$MFULastTextLength = 0;

$MFUWordwrapping = function(text, breite, höhe, noAdvice, endless) {
	var tl = 0, anzahl_Zeilen = [], maxcharat = null, charprüfung = false, textketten = [], maxtl = null, mintl = 0, prüfhöhe = 0, charWidth = 0, size = 0, lineHeight = 0, lastFont = 0;
	$MFULastTextHeight = 0;
	$MFULastTextLength = 0;
	if (breite == undefined || breite <= 36) {breite = SceneManager._screenWidth} // end of if
	breite -= 36;
	if (höhe == undefined || höhe <= 36) {höhe = SceneManager._screenHeight} // end of if
	höhe -= 36;
	var calcWindow = new Window_Base();
	while (true) {
		text = calcWindow.convertEscapeCharacters(text);
		size = calcWindow.contents.fontSize;
		lineHeight = calcWindow.contents.fontSize;
		lastFont = calcWindow.contents.fontSize;
		for (var i = 0; i <= text.length - 1; i++) {
			if(text.charCodeAt(i) === 27) {
				var skip = 0;
				while(true){
					if(text.charCodeAt(i+skip) === 93){
						if(text.charAt(i+1).toLowerCase() === 'i'){
							tl +=36;
							if(lineHeight < 32){
								lineHeight = 32;
							} // end of if
						} // end of if
						i += skip;
						break;
					} // end of if
					if(text.charAt(i+1) === '{'){
						lastFont = calcWindow.contents.fontSize;
						calcWindow.makeFontBigger();
						checkSize(i);
						i++;
						break;
					} // end of if
					if(text.charAt(i+1) === '}'){
						lastFont = calcWindow.contents.fontSize;
						calcWindow.makeFontSmaller();
						checkSize(i);
						i++;
						break;
					} // end of if
					if(skip > text.length){
						break;
					} // end of if
					skip++;
				} // end of while
			} else {
				charWidth = calcWindow.textWidth(text[i]);
				tl += charWidth;	
			} // end of if
			if (tl > breite || text.charCodeAt(i) == 10) {
				maxtl = tl;
				maxcharat = i;
				while (charprüfung == false) {
					if (i == mintl) {
						tl = maxtl;
						i = maxcharat;
						textketten_erstellen(i, 0);
					} else if (text.charAt(i) == ' ' || text.charAt(i) == undefined || text.charCodeAt(i) == 10) {
						textketten_erstellen(i, 1);
					} else if (text.charAt(i-1) == '-') {
						textketten_erstellen(i, 0);
					} else if (text.charAt(i) === '{') {
						i--;
						if(lastFont !== calcWindow.contents.fontSize){
							calcWindow.makeFontSmaller();
							lineHeight -= 12
						} // end of if
						textketten_erstellen(i, 0);
						i--;
					} else if (text.charAt(i) === '}') {
						i--;
						if(lastFont !== calcWindow.contents.fontSize){
							calcWindow.makeFontBigger();
						} // end of if
						textketten_erstellen(i, 0);
						i--;
					} else {
						i--;
						tl -= charWidth;
					} // end of if
				} // end of while
				$MFULastTextLength += tl;
				tl = 0;
				charprüfung = false;
			} // end of if
			if (i == text.length - 1) {
				textketten.push(text.substring(mintl, i + 1));
				calcHeight();
				$MFULastTextLength += tl;
			} // end of if
		} // end of for
		if (prüfhöhe > höhe && !endless) {
			if (size < 24) {
				if(noAdvice !== true){
					var fehlertext = "Der anzuzeigende Text übersteigt die festgelegte Höhe um " + (prüfhöhe - höhe) + " Pixel!";
					console.error(fehlertext);
					console.log(textketten);
					$MFUConsoleError();
				} // end of if
				break;
			} else {
				tl = 0; anzahl_Zeilen = []; charprüfung = false; textketten = []; mintl = 0; prüfhöhe = 0; $MFULastTextLength = 0;
				calcWindow.resetFontSettings();
				if(text.charCodeAt(0) === 27 && text.charAt(1) === '{'){
					text = text.substring(2);
				} else{
					text = "\\}" + text;
				} // end of if
			} // end of if
		} else {break} // end of if
	} // end of while
	text = "";
	for (var i = 0; i < textketten.length; i++) {
		if (i === textketten.length-1) {
			text = text.concat(textketten[i]);
		} else {
			text = text.concat(textketten[i], "\n");
		} // end of if
		$MFULastTextHeight += anzahl_Zeilen[i];
	} // end of for
	return text;
	function textketten_erstellen(i, mintl_plus) {
		textketten.push(text.substring(mintl, i));
		calcHeight();
		mintl = i + mintl_plus;
		charprüfung = true;
	} // end of textketten_erstellen
	function checkSize(n){
		if(lineHeight < calcWindow.contents.fontSize){
			lineHeight = calcWindow.contents.fontSize;
		} // end of if
		if(n === 0){
			size = calcWindow.contents.fontSize;
			lineHeight = calcWindow.contents.fontSize;
		} // end of if
	} // end of checkSize
	function calcHeight(){
		anzahl_Zeilen.push(lineHeight+8);
		prüfhöhe += lineHeight+8;
		lineHeight = calcWindow.contents.fontSize;
	} // end of calcHeight
} // end of Wordwrapping

// Zeigt ein Textfenster an
$MFUShowMessageWindow = function(text, hint, pos, face, faceIndex, scroll, noFast, interpreter) {
	if(typeof face === 'string' && $MFUIsNumber(faceIndex)){
		$gameMessage.setFaceImage(face,faceIndex);
	}else{
		$gameMessage.setFaceImage('',0);
	} // end of if
	if(scroll){
		$gameMessage.setScroll(scroll, noFast);
	} // end of if
	$gameMessage.setBackground(hint || 0)
	if (pos === undefined) {pos = 2} // end of if
	$gameMessage.setPositionType(pos)
	$gameMessage.add(text)
	if(interpreter){
		if(interpreter._waitMode !== undefined){
			interpreter.setWaitMode('message');
		} // end of if
	} // end of if
} // end of ShowMessageWindow

$MFULastIndex = 0;

// Prüft ob ein Objekt eine bestimmte Eigenschaft hat
function checkProperty(array, property, art) {
	for (var i = 0; i < array.length-1; i++) {
		if (array[i][property] == undefined) {
			console.error("Funktion '" + art + "' nicht möglich. Das Array muss aus Objekten bestehen die die Eigenschaft '" + property +"' haben müssen!");
			console.log(array[i]);
			$MFUConsoleError();
			return true;
		} // end of if
	} // end of for
	return false;
} // end of checkProperty

// Ermöglicht das sortieren von Arrays mit Objekten nach ihren Eigenschaften
function startSort(array, designator) {
	var nochmal = false;
	for (var i = 0; i < array.length-1; i++) {
		if (array[i][designator] > array[i+1][designator]) {
			array.splice(i, 2, array[i+1], array[i]);
			nochmal = true;
		} // end of if
	} // end of for
	if (nochmal) {startSort(array, designator)}
} // end of startSort
$MFUSort = function(array, designator) {
	designator = designator || 'bezeichnung';
	if(checkProperty(array, designator, '$MFUSort')){return}
	startSort(array, designator);
} // end of Sort

// Durchsucht ein Array nach Objekten deren Eigenschaft gleich einem bestimmten Begriff ist
$MFUIncludes = function(array, begriff, designator) {
	designator = designator || 'bezeichnung';
	if(checkProperty(array, designator, '$MFUIncludes')){return false}
	for (var i = 0; i < array.length; i++) {
		if (array[i][designator] == begriff) {
			$MFULastIndex = i;
			return true;
		} // end of if
	} // end of for
	return false;
} // end of Includes

// Wandelt arabische in römische Zahl um
$MFURomanNumerals = function(zahl) {
	var tsder = "", hdter = "", zehner = "", einer = "";
	while (zahl > 0) {
		if (zahl >= 1000) {
			tsder += "M";
			zahl -= 1000;
		} else if (zahl >= 500) {
			hdter += "D";
			zahl -= 500;
		} else if (zahl >= 100) {
			hdter += "C";
			zahl -= 100;
		} else if (zahl >= 50) {
			zehner += "L";
			zahl -= 50;
		} else if (zahl >= 10) {
			zehner += "X";
			zahl -= 10;
		} else if (zahl >= 5) {
			einer += "V";
			zahl -= 5;
		} else if (zahl >= 1) {
			einer += "I";
			zahl -= 1;
		} // end of if
		if (hdter == "DCCCC" ) {hdter = "CM"}
		if (hdter == "CCCC") {hdter = "CD"}
		if (zehner == "LXXXX") {zehner = "XC"}
		if (zehner == "XXXX") {zehner = "XL"}
		if (einer == "VIIII") {einer = "IX"}
		if (einer == "IIII") {einer = "IV"}
	} // end of while
	return tsder + hdter + zehner + einer
} // end of RomanNumerals

$MFUNumberIntoWord = function(zahl) {
	switch(zahl) {
		case 0 : zahl = "null"; break;
		case 1 : zahl = "eins"; break;
		case 2 : zahl = "zwei"; break;
		case 3 : zahl = "drei"; break;
		case 4 : zahl = "vier"; break;
		case 5 : zahl = "fünf"; break;
		case 6 : zahl = "sechs"; break;
		case 7 : zahl = "sieben"; break;
		case 8 : zahl = "acht"; break;
		case 9 : zahl = "neun"; break;
		case 10 : zahl = "zehn"; break;
		case 11 : zahl = "elf"; break;
		case 12 : zahl = "zwölf"; break;
	} // end of switch
	return zahl
} // end of NumberIntoWord

$MFUIsNumber = function (n) {
  return !isNaN(parseFloat(n)) && isFinite(n);
} // end of IsNumber

$MFUGainPropertyArray = function(arr, property){
	if(property === undefined){property = "bezeichnung"} // end of if
	if(checkProperty(arr, property, "$MFUGainPropertyArray")){
		return false;
	} else{
		var propArr = [];
		for (var i = 0; i < arr.length; i++) {
			propArr[i] = arr[i][property];
		} // end of for
		return propArr;
	} // end of if
} // end of GainPropertyArray

$MFUConsoleError = function(){
	alert("Bitte Konsole öffnen (F8 drücken), es gibt wichtige Meldungen!")
} // end of ConsoleError
$MFUAdvice = false;
})();